# Generated by Django 5.1.6 on 2025-08-11 04:47

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('saas_base', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='member',
            old_name='invite_email',
            new_name='email',
        ),
        migrations.AddField(
            model_name='member',
            name='name',
            field=models.CharField(blank=True, max_length=300, null=True, verbose_name='name'),
        ),
        migrations.AddField(
            model_name='member',
            name='is_staff',
            field=models.BooleanField(default=True),
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('name', models.CharField(max_length=100, primary_key=True, serialize=False)),
                ('description', models.CharField(blank=True, max_length=480)),
                ('created_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
                ('permissions', models.ManyToManyField(blank=True, to='saas_base.permission')),
            ],
            options={
                'verbose_name': 'role',
                'verbose_name_plural': 'roles',
                'db_table': 'saas_role',
                'ordering': ['created_at'],
            },
        ),
        migrations.AddField(
            model_name='member',
            name='role',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.SET_NULL,
                to='saas_base.role',
            ),
        ),
    ]



from __future__ import annotations

import os
import json
import functools
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from joserfc import jwe
from joserfc.errors import JoseError
from joserfc.jwk import OKPKey

DEFAULT_JWE_HEADER = {'alg': 'ECDH-ES+A256KW', 'enc': 'A128CBC-HS256'}


@functools.cache
def derive_key(secret: str) -> OKPKey:
    """Derive an X25519 key from a secret string and cache the result."""
    return OKPKey.derive_key(secret, 'X25519')


def encrypt_data(value: bytes) -> str:
    master_key = derive_key(settings.SECRET_KEY)
    header = getattr(settings, 'SAAS_JWE_HEADER', DEFAULT_JWE_HEADER)
    return jwe.encrypt_compact(
        header,
        plaintext=value,
        public_key=master_key,
    )


def decrypt_data(value: str) -> bytes | None:
    keys = [settings.SECRET_KEY] + settings.SECRET_KEY_FALLBACKS
    for key in keys:
        try:
            obj = jwe.decrypt_compact(value, private_key=derive_key(key))
            return obj.plaintext
        except JoseError:
            continue
    return None


@functools.cache
def load_secrets():
    secrets_file = getattr(settings, 'SAAS_SECRETS_FILE', None)
    if not secrets_file:
        raise ImproperlyConfigured('SAAS_SECRETS_FILE setting is not set')

    if not os.path.exists(secrets_file):
        return {}

    with open(secrets_file, 'r') as f:
        content = f.read().strip()

    if not content:
        return {}

    decrypted = decrypt_data(content)
    if decrypted is None:
        raise ImproperlyConfigured(f'Failed to decrypt secrets file: {secrets_file}')

    return json.loads(decrypted.decode('utf-8'))


def resolve_secret(key: str) -> str | None:
    if key.startswith('secret:'):
        return load_secrets()[key[7:]]
    return key



from django.test import override_settings
from saas_base.test import SaasTestCase
from saas_base.models import UserProfile


class TestUserAPI(SaasTestCase):
    user_id = SaasTestCase.GUEST_USER_ID

    def test_get_current_user(self):
        self.force_login()
        url = '/m/user/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsNone(data['avatar_url'])

        UserProfile.objects.create(user_id=self.user_id, avatar_url='https://example.com/avatar.jpg')
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['avatar_url'], 'https://example.com/avatar.jpg')

    def test_update_user(self):
        self.force_login()
        payload = {'first_name': 'First', 'avatar_url': 'https://example.com/avatar2.jpg'}
        resp = self.client.patch('/m/user/', data=payload)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['first_name'], 'First')
        self.assertEqual(data['avatar_url'], 'https://example.com/avatar2.jpg')

        payload = {'last_name': 'Last', 'avatar_url': 'https://example.com/avatar3.jpg'}
        resp = self.client.patch('/m/user/', data=payload)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['first_name'], 'First')
        self.assertEqual(data['last_name'], 'Last')
        self.assertEqual(data['avatar_url'], 'https://example.com/avatar3.jpg')

    def test_use_gravatar(self):
        self.force_login()
        resp = self.client.get('/m/user/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsNone(data['avatar_url'])

        with override_settings(SAAS={'ENABLE_GRAVATAR': True}):
            resp = self.client.get('/m/user/')
            self.assertEqual(resp.status_code, 200)
            data = resp.json()
            self.assertTrue(data['avatar_url'].startswith('https://gravatar.com/avatar/'))



from email.utils import formataddr
from django.utils.translation import gettext_lazy as _
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.mixins import (
    ListModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
)
from saas_base.drf.views import TenantEndpoint
from saas_base.drf.decorators import resource_permissions
from saas_base.drf.filters import TenantIdFilter, IncludeFilter, ChoiceFilter
from saas_base.serializers.member import (
    MemberSerializer,
    MemberInviteSerializer,
    MemberDetailSerializer,
)
from saas_base.models import Member
from saas_base.settings import saas_settings
from saas_base.signals import member_invited, mail_queued

__all__ = [
    'MemberListEndpoint',
    'MemberItemEndpoint',
]


class MemberListEndpoint(ListModelMixin, TenantEndpoint):
    email_template_id = 'invite_member'
    email_subject = _("You've Been Invited to Join %s")

    serializer_class = MemberSerializer
    filter_backends = [TenantIdFilter, IncludeFilter, ChoiceFilter]
    queryset = Member.objects.all()

    resource_name = 'tenant'
    resource_scopes = ['tenant', 'tenant:member']

    choice_filter_fields = ['status']
    include_select_related_fields = ['user', 'role']
    include_prefetch_related_fields = ['groups', 'permissions', 'groups__permissions']

    def get_email_subject(self):
        return self.email_subject % str(self.request.tenant)

    def get(self, request: Request, *args, **kwargs):
        """List all members in the tenant."""
        return self.list(request, *args, **kwargs)

    @resource_permissions('tenant.admin')
    def post(self, request: Request, *args, **kwargs):
        """Invite a member to join the tenant."""
        tenant_id = self.get_tenant_id()
        context = self.get_serializer_context()
        serializer = MemberInviteSerializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        member = serializer.save(tenant_id=tenant_id, inviter=request.user)

        member_invited.send(self.__class__, member=member, request=request)
        if member.status != Member.InviteStatus.ACTIVE:
            if member.name:
                recipient = formataddr((member.name, member.email))
            else:
                recipient = member.email

            invite_link = saas_settings.MEMBER_INVITE_LINK % str(member.id)
            if not invite_link.startswith('http'):
                invite_link = request.build_absolute_uri(invite_link)

            mail_queued.send(
                sender=self.__class__,
                template_id=self.email_template_id,
                subject=str(self.get_email_subject()),
                recipients=[recipient],
                context={
                    'inviter': request.user,
                    'member': member,
                    'tenant': request.tenant,
                    'invite_link': invite_link,
                },
                request=request,
            )
        data = MemberDetailSerializer(member).data
        return Response(data)


class MemberItemEndpoint(UpdateModelMixin, DestroyModelMixin, TenantEndpoint):
    serializer_class = MemberDetailSerializer
    queryset = Member.objects.all()
    resource_name = 'tenant'
    resource_scopes = ['tenant', 'tenant:member']

    def get(self, request: Request, *args, **kwargs):
        """Retrieve the information of a member."""
        queryset = self.filter_queryset(self.get_queryset())
        queryset = queryset.prefetch_related('groups', 'permissions', 'groups__permissions')
        member = self.get_object_or_404(queryset, pk=kwargs['pk'])
        self.check_object_permissions(request, member)
        serializer = self.get_serializer(member)
        return Response(serializer.data)

    @resource_permissions('tenant.admin')
    def patch(self, request: Request, *args, **kwargs):
        """Update a member's permissions and groups."""
        return self.partial_update(request, *args, **kwargs)

    @resource_permissions('tenant.admin')
    def delete(self, request: Request, *args, **kwargs):
        """Remove a member from the tenant."""
        return self.destroy(request, *args, **kwargs)



import uuid
from typing import List, Set
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.utils.functional import cached_property
from .permission import Permission
from .role import Role
from .group import Group
from ..settings import saas_settings
from ..db import CachedManager


class MemberManager(CachedManager['Member']):
    natural_key = ['tenant_id', 'user_id']
    query_select_related = ['user', 'role']

    def get_by_natural_key(self, tenant_id, user_id) -> 'Member':
        return self.get_from_cache_by_natural_key(tenant_id, user_id)


class Member(models.Model):
    class InviteStatus(models.IntegerChoices):
        REQUEST = 0, 'request'
        WAITING = 1, 'waiting'
        ACTIVE = 2, 'active'

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    tenant = models.ForeignKey(settings.SAAS_TENANT_MODEL, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, blank=True, null=True)

    name = models.CharField(_('name'), max_length=300, null=True, blank=True)
    email = models.EmailField(null=True, blank=True)

    # this member is invited by who
    inviter = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='+',
    )

    status = models.SmallIntegerField(default=InviteStatus.REQUEST, choices=InviteStatus.choices)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    role = models.ForeignKey(Role, on_delete=models.SET_NULL, blank=True, null=True)
    groups = models.ManyToManyField(
        Group,
        blank=True,
        help_text=_(
            'The groups this user belongs to. A user will get all permissions granted to each of their groups.'
        ),
    )
    permissions = models.ManyToManyField(
        Permission,
        blank=True,
        help_text=_('Specific permissions for this user.'),
    )
    objects = MemberManager()

    class Meta:
        verbose_name = _('member')
        verbose_name_plural = _('members')
        unique_together = [
            ['tenant', 'user'],
        ]
        ordering = ['-created_at']
        db_table = 'saas_member'

    def __str__(self):
        if self.user:
            return str(self.user)
        if self.email:
            return self.email
        return ''

    def natural_key(self):
        return self.tenant_id, self.user_id

    @property
    def is_active(self) -> bool:
        return self.status == self.InviteStatus.ACTIVE

    @cached_property
    def group_permissions(self) -> List[str]:
        return self.__get_group_permissions()

    @cached_property
    def role_permissions(self) -> List[str]:
        return self.__get_role_permissions()

    @cached_property
    def user_permissions(self) -> List[str]:
        return self.__get_user_permissions()

    def get_all_permissions(self) -> Set[str]:
        perms = set()
        if 'permissions' in saas_settings.MEMBER_PERMISSION_MANAGERS:
            perms = perms.union(self.user_permissions)
        if 'role' in saas_settings.MEMBER_PERMISSION_MANAGERS:
            perms = perms.union(self.role_permissions)
        if 'groups' in saas_settings.MEMBER_PERMISSION_MANAGERS:
            perms = perms.union(self.group_permissions)
        return perms

    def __get_group_permissions(self):
        return list(self.groups.values_list('permissions__name', flat=True))

    def __get_role_permissions(self):
        if self.role_id:
            return list(self.role.permissions.all().values_list('name', flat=True))
        return []

    def __get_user_permissions(self):
        return list(self.permissions.all().values_list('name', flat=True))



from django.urls import path
from ..endpoints.password import (
    PasswordResetEndpoint,
    PasswordForgotEndpoint,
)
from ..endpoints.auth import (
    LogoutEndpoint,
    PasswordLogInEndpoint,
    SignupRequestEndpoint,
    SignupConfirmEndpoint,
    SignupWithInvitationEndpoint,
    InvitationEndpoint,
)


urlpatterns = [
    path('logout/', LogoutEndpoint.as_view()),
    path('login/', PasswordLogInEndpoint.as_view()),
    path('signup/request/', SignupRequestEndpoint.as_view()),
    path('signup/confirm/', SignupConfirmEndpoint.as_view()),
    path('signup/via/<pk>/', SignupWithInvitationEndpoint.as_view()),
    path('password/forgot/', PasswordForgotEndpoint.as_view()),
    path('password/reset/', PasswordResetEndpoint.as_view()),
    path('invitation/<pk>/', InvitationEndpoint.as_view()),
]



from django.contrib import admin
from .models import (
    Permission,
    Role,
    Group,
    Member,
    UserEmail,
    Tenant,
)


@admin.register(Permission)
class PermissionAdmin(admin.ModelAdmin):
    list_display = ['name', 'internal', 'created_at']


@admin.register(Role)
class RoleAdmin(admin.ModelAdmin):
    list_display = ['name', 'description', 'created_at']


@admin.register(Tenant)
class TenantAdmin(admin.ModelAdmin):
    list_display = ['id', 'name', 'slug', 'owner', 'expires_at', 'created_at']


@admin.register(Group)
class GroupAdmin(admin.ModelAdmin):
    list_display = ['id', 'tenant', 'name', 'managed', 'created_at']


@admin.register(Member)
class MemberAdmin(admin.ModelAdmin):
    list_display = ['id', 'tenant', 'user', 'name', 'email', 'status', 'created_at']


@admin.register(UserEmail)
class UserEmailAdmin(admin.ModelAdmin):
    list_display = ['id', 'email', 'primary', 'verified', 'created_at']



from django.urls import path
from ..endpoints.tenant import (
    TenantListEndpoint,
    TenantItemEndpoint,
)

urlpatterns = [
    path('', TenantListEndpoint.as_view()),
    path('<pk>/', TenantItemEndpoint.as_view()),
]



# Generated by Django 6.0.1 on 2026-01-11 05:52

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saas_base', '0004_tenant_logo'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='profile', serialize=False, to=settings.AUTH_USER_MODEL)),
                ('bio', models.TextField(blank=True, max_length=500)),
                ('birth_date', models.DateField(blank=True, null=True)),
                ('website', models.URLField(blank=True, null=True)),
                ('avatar_url', models.URLField(blank=True, null=True)),
                ('region', models.CharField(blank=True, max_length=4, null=True)),
                ('locale', models.CharField(blank=True, max_length=10, null=True)),
                ('timezone', models.CharField(default='UTC', help_text="e.g., 'America/New_York'", max_length=50)),
            ],
            options={
                'db_table': 'saas_user_profile',
            },
        ),
        migrations.AlterModelTable(
            name='useremail',
            table='saas_user_email',
        ),
    ]



from rest_framework import serializers
from ..models import Permission


class PermissionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Permission
        fields = ['name', 'description', 'internal']



from django.urls import path
from ..endpoints.groups import (
    GroupListEndpoint,
    GroupItemEndpoint,
)

urlpatterns = [
    path('', GroupListEndpoint.as_view()),
    path('<pk>/', GroupItemEndpoint.as_view()),
]



import functools


def resource_permissions(*names):
    """A decorator that adding required resource permissions to the view function.

    .. code-block:: python

        class TestView(Endpoint):
            @resource_permissions('tenant.admin')
            def post(self, request, *args, **kwargs):
                ...
    """

    def decorated(func):
        func._resource_permissions = names

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

    return decorated



from typing import TYPE_CHECKING
from drf_spectacular.extensions import OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import build_array_type

if TYPE_CHECKING:
    from drf_spectacular.openapi import AutoSchema
    from drf_spectacular.utils import Direction


class ChoiceFieldFix(OpenApiSerializerFieldExtension):
    target_class = 'saas_base.drf.serializers.ChoiceField'

    def map_serializer_field(self, auto_schema: 'AutoSchema', direction: 'Direction'):
        choices = list(self.target.choices.values())
        return {'type': 'string', 'enum': choices}


class RelatedSerializerField(OpenApiSerializerFieldExtension):
    target_class = 'saas_base.drf.serializers.RelatedSerializerField'

    def map_serializer_field(self, auto_schema: 'AutoSchema', direction: 'Direction'):
        if direction == 'response':
            schema = auto_schema._map_serializer(self.target.serializer_cls, direction, bypass_extensions=True)
        else:
            model = self.target.serializer_cls.Meta.model
            pk_field = model._meta.pk
            schema = auto_schema._map_model_field(pk_field, direction)

        if self.target.many:
            schema = build_array_type(schema)
        return schema



from django.utils.translation import gettext as _
from django.contrib.auth import password_validation, authenticate
from django.contrib.auth.models import AbstractUser
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from .email_code import (
    EmailCodeRequestSerializer,
    EmailCodeConfirmSerializer,
    RetrieveUserEmailMixin,
)
from ..models import UserEmail

PASSWORD_CODE = 'saas:password_code'


class PasswordLoginSerializer(serializers.Serializer):
    username = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    @staticmethod
    def invalid_errors():
        errors = {'password': [_('Invalid username or password.')]}
        return ValidationError(errors)

    def create(self, validated_data):
        request = self.context['request']
        user = authenticate(request=request, **validated_data)
        if not user:
            raise self.invalid_errors()
        return user

    def update(self, instance, validated_data):
        raise RuntimeError('This method is not allowed.')


class PasswordForgetSerializer(RetrieveUserEmailMixin, EmailCodeRequestSerializer):
    CACHE_PREFIX = PASSWORD_CODE

    def create(self, validated_data) -> UserEmail:
        user_email = validated_data['email']
        return user_email


class PasswordResetSerializer(RetrieveUserEmailMixin, EmailCodeConfirmSerializer):
    CACHE_PREFIX = PASSWORD_CODE
    password = serializers.CharField(required=True)

    def validate_password(self, raw_password):
        password_validation.validate_password(raw_password)
        return raw_password

    def create(self, validated_data):
        obj: UserEmail = validated_data['email']
        user_id = validated_data['code']

        if not user_id or obj.user_id != user_id:
            raise ValidationError({'code': [_('Code does not match or expired.')]})

        raw_password = validated_data['password']
        user: AbstractUser = obj.user
        user.set_password(raw_password)
        user.save()
        return obj



from django.utils.translation import gettext_lazy as _
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.mixins import (
    ListModelMixin,
    DestroyModelMixin,
    RetrieveModelMixin,
)
from rest_framework.throttling import UserRateThrottle
from saas_base.drf.views import AuthenticatedEndpoint
from saas_base.drf.filters import CurrentUserFilter
from saas_base.models import UserEmail
from saas_base.serializers.user_email import (
    UserEmailSerializer,
    AddEmailRequestSerializer,
    AddEmailConfirmSerializer,
)
from saas_base.signals import mail_queued

__all__ = [
    'UserEmailListEndpoint',
    'UserEmailItemEndpoint',
    'AddUserEmailRequestEndpoint',
    'AddUserEmailConfirmEndpoint',
]


class UserEmailListEndpoint(ListModelMixin, AuthenticatedEndpoint):
    resource_scopes = ['user:email']
    pagination_class = None
    serializer_class = UserEmailSerializer
    queryset = UserEmail.objects.all()
    filter_backends = [CurrentUserFilter]

    def get(self, request: Request, *args, **kwargs):
        """List all the current user's emails."""
        return self.list(request, *args, **kwargs)


class UserEmailItemEndpoint(RetrieveModelMixin, DestroyModelMixin, AuthenticatedEndpoint):
    resource_scopes = ['user:email']
    pagination_class = None
    serializer_class = UserEmailSerializer
    queryset = UserEmail.objects.all()
    filter_backends = [CurrentUserFilter]

    def get(self, request: Request, *args, **kwargs):
        """Retrieve the current user's emails with the given uuid.'"""
        return self.retrieve(request, *args, **kwargs)

    def patch(self, request: Request, *args, **kwargs):
        """Set this email to be the primary email address."""
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)

    def delete(self, request: Request, *args, **kwargs):
        """List all the current user's emails."""
        return self.destroy(request, *args, **kwargs)


class AddUserEmailRequestEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user:email']
    email_template_id = 'add_email'
    email_subject = _('Add new email')

    throttle_classes = [UserRateThrottle]
    serializer_class = AddEmailRequestSerializer

    def post(self, request: Request):
        """Send a request of authorization code for linking the account."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()
        mail_queued.send(
            sender=self.__class__,
            template_id=self.email_template_id,
            subject=self.email_subject,
            recipients=[obj.recipient()],
            context={'code': obj.code, 'user': obj.user},
            request=request,
        )
        return Response(status=204)


class AddUserEmailConfirmEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user:email']
    throttle_classes = [UserRateThrottle]
    serializer_class = AddEmailConfirmSerializer

    def post(self, request: Request):
        """Reset password of a user with the given code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()
        data = UserEmailSerializer(obj).data
        return Response(data)



from django.conf import settings
from django.core.signals import setting_changed
from django.utils.module_loading import import_string


DEFAULTS = {
    'SITE': {
        'name': 'Django SaaS',
        'url': 'https://django-saas.dev',
        'icon': '',
        'copyright': '© 2025',
    },
    'ENABLE_GRAVATAR': False,
    'GRAVATAR_PARAMS': {'d': 'identicon', 'r': 'g', 's': 200},
    'CLIENT_IP_HEADERS': None,
    'DB_CACHE_ALIAS': 'default',
    'PERMISSION_NAME_FORMATTER': '{resource}.{action}',
    'DEFAULT_REGION': '',
    'TENANT_ID_HEADER': 'X-Tenant-Id',
    'DEFAULT_FROM_EMAIL': None,
    'LOGIN_SECURITY_RULES': [],
    'SIGNUP_SECURITY_RULES': [],
    'RESET_PASSWORD_SECURITY_RULES': [],
    'SIGNUP_REQUEST_CREATE_USER': False,
    'MEMBER_INVITE_LINK': '/invite/%s/',
    'MEMBER_PERMISSION_MANAGERS': ['permissions', 'groups', 'role'],
}


class Settings:
    IMPORT_PROVIDERS = [
        'MAIL_PROVIDERS',
        'LOGIN_SECURITY_RULES',
        'SIGNUP_SECURITY_RULES',
        'RESET_PASSWORD_SECURITY_RULES',
    ]

    def __init__(self, settings_key='SAAS', user_settings=None, defaults=None):
        self.settings_key = settings_key
        self._user_settings = user_settings
        if defaults is None:
            defaults = DEFAULTS
        self.defaults = defaults
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if self._user_settings is None:
            self._user_settings = getattr(settings, self.settings_key, {})
        return self._user_settings

    def __getitem__(self, attr):
        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]
        return val

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid %s setting: '%s'" % (self.settings_key, attr))

        val = self[attr]

        # Coerce import strings into classes
        if attr in self.IMPORT_PROVIDERS:
            val = perform_import(val)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self, value=None):
        for attr in self._cached_attrs:
            delattr(self, attr)

        self._cached_attrs.clear()
        if isinstance(value, dict):
            self.user_settings.update(value)
        else:
            self._user_settings = None

    def listen_setting_changed(self, setting, **kwargs):
        if setting == self.settings_key:
            self.reload(kwargs.get('value'))


def perform_import_provider(data):
    backend_cls = import_string(data['backend'])
    options = data.get('options', {})
    return backend_cls(**options)


def perform_import(val):
    if val is None:
        return None

    elif isinstance(val, (list, tuple)):
        return [perform_import_provider(item) for item in val]
    elif isinstance(val, dict) and 'backend' not in val:
        return {k: perform_import_provider(val[k]) for k in val}
    else:
        return perform_import_provider(val)


saas_settings = Settings()
setting_changed.connect(saas_settings.listen_setting_changed)



from django.core import mail
from django.test import override_settings
from saas_base.models import UserEmail, Member
from tests.client import FixturesTestCase


class TestSignUpWithoutCreateUser(FixturesTestCase):
    user_id = FixturesTestCase.ADMIN_USER_ID

    def test_signup_success(self):
        data1 = {'username': 'demo', 'email': 'hi@foo.com', 'password': 'hello world'}
        resp = self.client.post('/m/auth/signup/request/', data=data1)
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(len(mail.outbox), 1)
        data2 = {**data1, 'code': self.get_mail_auth_code()}
        resp = self.client.post('/m/auth/signup/confirm/', data=data2)
        self.assertEqual(resp.status_code, 200)

    def test_signup_existed_email(self):
        user = self.get_user()
        UserEmail.objects.create(user=user, email='hi@foo.com', primary=True, verified=True)
        data = {'username': 'foo', 'email': 'hi@foo.com', 'password': 'hello world'}
        resp = self.client.post('/m/auth/signup/request/', data=data)
        self.assertEqual(resp.status_code, 400)
        self.assertIn('existing', resp.json()['email'][0])

    def test_signup_blocked_email(self):
        rules = [{'backend': 'saas_base.rules.BlockedEmailDomains'}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {'username': 'bar', 'email': 'hi@boofx.com', 'password': 'hello world'}
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 400)

        rules = [{'backend': 'saas_base.rules.BlockedEmailDomains', 'options': {'domains': ['bar.com']}}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {'username': 'bar', 'email': 'hi@bar.com', 'password': 'hello world'}
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 400)

    def test_signup_too_many_dots(self):
        rules = [{'backend': 'saas_base.rules.AvoidTooManyDots'}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {'username': 'bar', 'email': 'a.b.c.d.e.f@bar.com', 'password': 'hello world'}
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 400)

    def test_signup_using_plus(self):
        rules = [{'backend': 'saas_base.rules.AvoidUsingPlus'}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {'username': 'bar', 'email': 'username+demo@gmail.com', 'password': 'hello world'}
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 400)

    def test_signup_turnstile(self):
        rules = [{'backend': 'saas_base.rules.Turnstile'}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {'username': 'bar', 'email': 'hi@bar.com', 'password': 'hello world'}
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 400)

            data = {**data, 'cf-turnstile-response': '**token**'}
            with self.mock_requests('turnstile_success.json'):
                resp = self.client.post('/m/auth/signup/request/', data=data)
                self.assertEqual(resp.status_code, 204)

            data = {**data, 'cf-turnstile-response': '**token**'}
            with self.mock_requests('turnstile_failed.json'):
                resp = self.client.post('/m/auth/signup/request/', data=data)
                self.assertEqual(resp.status_code, 400)

    def test_turnstile_with_real_http(self):
        rules = [{'backend': 'saas_base.rules.Turnstile', 'options': {'secret': 'secret:turnstile'}}]
        with override_settings(SAAS={'SIGNUP_SECURITY_RULES': rules}):
            data = {
                'username': 'bar',
                'email': 'hi@bar.com',
                'password': 'hello world',
                'cf-turnstile-response': '**token**',
            }
            resp = self.client.post('/m/auth/signup/request/', data=data)
            self.assertEqual(resp.status_code, 204)

    def test_signup_with_membership_invite(self):
        # prepare membership
        email = 'hi@foo.com'
        Member.objects.create(tenant_id=self.tenant_id, email=email)

        data1 = {'username': 'demo', 'email': email, 'password': 'hello world'}
        self.client.post('/m/auth/signup/request/', data=data1)
        data2 = {**data1, 'code': self.get_mail_auth_code()}
        resp = self.client.post('/m/auth/signup/confirm/', data=data2)
        self.assertEqual(resp.status_code, 200)
        obj = UserEmail.objects.get(email=email)
        member = Member.objects.get(user_id=obj.user_id, tenant_id=self.tenant_id)
        self.assertEqual(member.status, Member.InviteStatus.WAITING)

    def test_signup_via_member_invite(self):
        email = 'hi@foo.com'
        obj = Member.objects.create(tenant_id=self.tenant_id, email=email)
        data = {'username': 'demo', 'password': 'hello world'}
        resp = self.client.post(f'/m/auth/signup/via/{obj.pk}/', data=data)
        self.assertEqual(resp.status_code, 200)

    def test_signup_with_waiting_member(self):
        user = self.get_user()
        obj = Member.objects.create(
            tenant_id=self.tenant_id,
            user=user,
            email=user.email,
            status=Member.InviteStatus.WAITING,
        )
        data = {'username': 'demo', 'password': 'hello world'}
        resp = self.client.post(f'/m/auth/signup/via/{obj.pk}/', data=data)
        self.assertEqual(resp.status_code, 404)


@override_settings(SAAS={'SIGNUP_REQUEST_CREATE_USER': True})
class TestSignUpWithCreateUser(FixturesTestCase):
    def test_signup_success(self):
        data1 = {'username': 'demo', 'email': 'hi@foo.com', 'password': 'hello world'}
        resp = self.client.post('/m/auth/signup/request/', data=data1)
        self.assertEqual(resp.status_code, 204)
        obj = UserEmail.objects.get(email='hi@foo.com')
        self.assertEqual(obj.verified, False)

        data2 = {'code': self.get_mail_auth_code(), 'email': 'hi@foo.com'}
        resp = self.client.post('/m/auth/signup/confirm/', data=data2)
        self.assertEqual(resp.status_code, 200)
        obj = UserEmail.objects.get(email='hi@foo.com')
        self.assertEqual(obj.verified, True)


class TestLoginAPI(FixturesTestCase):
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_login_with_username(self):
        user = self.get_user()

        data = {'username': user.username, 'password': 'hello world'}
        resp = self.client.post('/m/auth/login/', data=data)
        self.assertEqual(resp.status_code, 400)

        user.set_password('hello world')
        user.save()

        resp = self.client.post('/m/auth/login/', data=data)
        self.assertEqual(resp.status_code, 200)
        self.assertIn('next', resp.json())

    def test_login_with_email(self):
        user = self.get_user()

        data = {'username': 'hi@foo.com', 'password': 'hello world'}
        resp = self.client.post('/m/auth/login/', data=data)
        self.assertEqual(resp.status_code, 400)

        user.set_password('hello world')
        user.save()

        obj = UserEmail.objects.create(user=user, email='hi@foo.com')
        resp = self.client.post('/m/auth/login/', data=data)
        self.assertEqual(resp.status_code, 400)

        obj.primary = True
        obj.verified = True
        obj.save()

        resp = self.client.post('/m/auth/login/', data=data)
        self.assertEqual(resp.status_code, 200)

        resp = self.client.get('/m/user/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['username'], user.username)


class TestInvitationAPI(FixturesTestCase):
    user_id = FixturesTestCase.ADMIN_USER_ID

    def test_request_status_invitation(self):
        email = 'hi@foo.com'
        obj = Member.objects.create(tenant_id=self.tenant_id, email=email)
        resp = self.client.get(f'/m/auth/invitation/{obj.pk}/')
        self.assertEqual(resp.status_code, 200)
        result = resp.json()
        self.assertEqual(result['status'], 'request')

    def test_waiting_status_invitation(self):
        user = self.get_user(self.EMPTY_USER_ID)
        obj = Member.objects.create(
            tenant_id=self.tenant_id,
            inviter_id=self.user_id,
            user=user,
            email=user.email,
            status=Member.InviteStatus.WAITING,
        )
        resp = self.client.get(f'/m/auth/invitation/{obj.pk}/')
        self.assertEqual(resp.status_code, 200)
        result = resp.json()
        self.assertEqual(result['status'], 'waiting')

    def test_active_status_invitation(self):
        user = self.get_user(self.EMPTY_USER_ID)
        obj = Member.objects.create(
            tenant_id=self.tenant_id,
            inviter_id=self.user_id,
            user=user,
            email=user.email,
            status=Member.InviteStatus.ACTIVE,
        )
        resp = self.client.get(f'/m/auth/invitation/{obj.pk}/')
        self.assertEqual(resp.status_code, 404)



import pytest
from django.test import override_settings
from django.core.management import call_command
from django.core.management.base import CommandError


@override_settings(SAAS_SECRETS_FILE=None)
def test_no_secrets_file():
    with pytest.raises(CommandError, match="SAAS_SECRETS_FILE setting is not set"):
        call_command('secrets', 'list')


@override_settings(SAAS_SECRETS_FILE='.not-exists-secrets')
def test_secrets_list_empty(capsys):
    call_command('secrets', 'list')
    captured = capsys.readouterr()
    assert "No secrets found." in captured.out


def test_list_default_secrets(capsys):
    call_command('secrets', 'list')
    captured = capsys.readouterr()
    assert "turnstile" in captured.out



from django.apps import AppConfig


class DemoConfig(AppConfig):
    name = 'tests.demo_app'

    def ready(self):
        from django.utils.module_loading import autodiscover_modules
        autodiscover_modules('event_hooks')



from django.db import connection
from django.test import override_settings
from saas_base.test import SaasTestCase
from tests.demo_app.models import UserSecret


class TestUserSecrets(SaasTestCase):
    user_id = SaasTestCase.GUEST_USER_ID

    def get_db_value(self, field: str, model_id: int):
        cursor = connection.cursor()
        cursor.execute(
            f"select {field} from demo_app_usersecret where id = {model_id};"
        )
        return cursor.fetchone()[0]

    def test_create_user_secrets(self):
        obj = UserSecret(secret_key=b'test')
        obj.save()
        self.assertEqual(obj.secret_key, b'test')

        db_value = self.get_db_value('secret_key', obj.pk)
        self.assertEqual(len(db_value.split('.')), 5)

    def test_key_fallback(self):
        obj = UserSecret(secret_key=b'test')
        obj.save()
        self.assertEqual(obj.secret_key, b'test')

        with override_settings(SECRET_KEY='new-primary-key'):
            new_obj = UserSecret.objects.get(pk=obj.pk)
            # we cannot decreypt it
            self.assertIsNone(new_obj.secret_key)

        with override_settings(SECRET_KEY='new-primary-key', SECRET_KEY_FALLBACKS=['django-insecure']):
            fallback_obj = UserSecret.objects.get(pk=obj.pk)
            self.assertEqual(fallback_obj.secret_key, b'test')



import random
import string
from email.utils import formataddr
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..models import UserEmail


ERRORS = {
    'code': _('Code does not match or expired.'),
    'exist_email': _('This email address is already associated with an existing account.'),
    'invalid_email': _('This email address is not associated with your account.'),
}


class EmailCode:
    def __init__(self, email: str, code: str, user=None):
        self.email = email
        self.code = code
        self.user = user

    def recipient(self):
        if self.user:
            return formataddr((self.user.username, self.email))
        return self.email


class EmailCodeRequestSerializer(serializers.Serializer):
    CACHE_PREFIX = 'saas:email_code'
    email = serializers.EmailField(required=True)

    def save_auth_code(self, value):
        email = self.validated_data['email']
        code = ''.join(random.sample(string.ascii_uppercase, 6))
        cache_key = f'{self.CACHE_PREFIX}:{email}:{code}'
        cache.set(cache_key, value, timeout=300)
        return code


class EmailCodeConfirmSerializer(serializers.Serializer):
    CACHE_PREFIX = 'saas:email_code'
    email = serializers.EmailField(required=True)
    code = serializers.CharField(required=True, max_length=6)

    def fail_code(self):
        raise ValidationError(ERRORS['code'])

    def validate_code(self, code: str):
        email = self.initial_data['email']
        code = code.upper()
        cache_key = f'{self.CACHE_PREFIX}:{email}:{code}'
        value: str = cache.get(cache_key)
        if not value:
            self.fail_code()

        cache.delete(cache_key)
        return value


class NewUserEmailMixin:
    def validate_email(self, email: str):
        email = UserEmail.objects.normalize_email(email)
        try:
            UserEmail.objects.get(email=email)
            raise ValidationError(ERRORS['exist_email'])
        except UserEmail.DoesNotExist:
            return email


class RetrieveUserEmailMixin(serializers.Serializer):
    def validate_email(self, email: str):
        email = UserEmail.objects.normalize_email(email)
        try:
            user_email = UserEmail.objects.select_related('user').get(email=email)
        except UserEmail.DoesNotExist:
            raise ValidationError(ERRORS['invalid_email'])
        return user_email



from tests.client import FixturesTestCase
from saas_base.models import Member


class TestMembersAPI(FixturesTestCase):
    user_id = FixturesTestCase.GUEST_USER_ID

    def test_list_users_via_owner(self):
        self.force_login(self.OWNER_USER_ID)
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 200)

    def test_list_users_via_guest_user(self):
        self.force_login()
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 403)

        self.add_user_perms('tenant.read')
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertNotIn('user', member)

    def test_list_include_user(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        resp = self.client.get('/m/members/?include=user')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertIn('user', member)

    def test_list_include_permissions(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        resp = self.client.get('/m/members/?include=permissions')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertIn('permissions', member)

    def test_list_members_by_status(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        resp = self.client.get('/m/members/?status=request')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['count'], 0)
        resp = self.client.get('/m/members/?status=active')
        data = resp.json()
        self.assertNotEqual(data['count'], 0)

    def test_invite_member_signup(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        data = {'name': 'Django'}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json(), {'email': ['This field is required.']})

        # invite non-exist user
        data = {'email': 'signup@example.com'}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        member = resp.json()
        self.assertEqual(member['status'], 'request')

        # invite existing user
        user = self.get_user(self.STAFF_USER_ID)
        data = {'email': user.email}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        member = resp.json()
        self.assertEqual(member['status'], 'waiting')

    def test_invite_with_permissions(self):
        self.add_user_perms('tenant.admin')
        self.force_login()

        user = self.get_user(self.STAFF_USER_ID)
        data = {'email': user.email, 'permissions': ['tenant.read']}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        permissions = resp.json()['permissions']
        self.assertEqual(permissions[0]['name'], 'tenant.read')

    def test_invite_with_role(self):
        self.add_user_perms('tenant.admin')
        self.force_login()

        user = self.get_user(self.STAFF_USER_ID)
        data = {'email': user.email, 'role': 'admin'}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        permissions = [p['name'] for p in resp.json()['role']['permissions']]
        self.assertIn('tenant.read', permissions)

    def test_view_member_item(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.get(f'/m/members/{member.id}/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIn('groups', data)
        self.assertIn('permissions', data)

    def test_update_member_item(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        data = {'permissions': ['tenant.read']}
        resp = self.client.patch(f'/m/members/{member.id}/', data=data)
        self.assertEqual(resp.status_code, 403)
        self.add_user_perms('tenant.admin')
        resp = self.client.patch(f'/m/members/{member.id}/', data=data)
        self.assertEqual(resp.status_code, 200)
        permissions = resp.json()['permissions']
        result = [p['name'] for p in permissions]
        self.assertEqual(result, ['tenant.read'])

    def test_remove_member_item(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.delete(f'/m/members/{member.id}/')
        self.assertEqual(resp.status_code, 204)

    def test_invite_duplicate_member(self):
        self.force_login(self.OWNER_USER_ID)
        email = 'demo-1@example.com'
        data = {'email': email, 'permissions': ['tenant.read']}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 400)
        data = resp.json()
        self.assertEqual(data['email'], ['This user has already been invited.'])

    def test_already_invited(self):
        self.force_login(self.OWNER_USER_ID)
        member = Member.objects.filter(
            tenant_id=self.tenant_id,
            user_id=self.OWNER_USER_ID,
        ).first()
        member.email = 'demo-1@example.com'
        member.save()
        data = {'email': member.email, 'permissions': ['tenant.read']}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 400)
        data = resp.json()
        self.assertEqual(data['email'], ['This email has already been invited.'])

    def test_invite_self(self):
        self.force_login(self.OWNER_USER_ID)
        # clean self member
        Member.objects.filter(
            tenant_id=self.tenant_id,
            user_id=self.OWNER_USER_ID,
        ).delete()
        email = 'demo-1@example.com'
        data = {'email': email, 'permissions': ['tenant.read']}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['status'], 'active')



from __future__ import annotations

import hashlib
from urllib.parse import urlencode


def gen_gravatar_url(email: str, name: str | None = None, **params):
    email_sha = hashlib.sha256(email.encode("utf-8")).hexdigest()
    url = f"https://gravatar.com/avatar/{email_sha}"
    if name:
        params.update({'d': 'initials', 'name': name})
    params.setdefault('s', 200)
    return f"{url}?{urlencode(params)}"



from rest_framework import serializers
from ..models import get_tenant_model
from ..signals import (
    before_create_tenant,
    before_update_tenant,
)


class TenantSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_tenant_model()
        exclude = ['owner']
        read_only_fields = ['id', 'expires_at', 'created_at', 'updated_at']

    def create(self, validated_data):
        before_create_tenant.send(self.__class__, data=validated_data, **self.context)
        return super().create(validated_data)

    def update(self, instance, validated_data):
        before_update_tenant.send(self.__class__, tenant=instance, data=validated_data, **self.context)
        return super().update(instance, validated_data)


class TenantUpdateSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_tenant_model()
        exclude = ['owner']
        read_only_fields = [
            'id',
            'slug',
            'environment',
            'region',
            'expires_at',
            'created_at',
            'updated_at',
        ]

    def update(self, instance, validated_data):
        before_update_tenant.send(self.__class__, tenant=instance, data=validated_data, **self.context)
        return super().update(instance, validated_data)



from django.utils.translation import gettext as _
from django.db import IntegrityError
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..drf.serializers import ModelSerializer, RelatedSerializerField
from ..models import Member, UserEmail
from .tenant import TenantSerializer
from .permission import PermissionSerializer
from .role import RoleSerializer
from .group import GroupSerializer
from .user import UserSerializer, SimpleUserSerializer


class MemberSerializer(ModelSerializer):
    user = UserSerializer(required=False, read_only=True)
    inviter = SimpleUserSerializer(required=False, read_only=True)
    role = RoleSerializer(required=False, read_only=True)
    groups = GroupSerializer(required=False, many=True, read_only=True)
    permissions = PermissionSerializer(required=False, many=True, read_only=True)

    class Meta:
        model = Member
        exclude = ['tenant']
        request_include_fields = ['user', 'role', 'groups', 'permissions']


class MemberInviteSerializer(ModelSerializer):
    name = serializers.CharField(required=False, allow_blank=True)
    email = serializers.EmailField(required=True)
    permissions = RelatedSerializerField(PermissionSerializer, many=True, required=False)
    role = RelatedSerializerField(RoleSerializer, required=False)
    groups = RelatedSerializerField(GroupSerializer, many=True, required=False)

    class Meta:
        model = Member
        fields = ['name', 'email', 'permissions', 'role', 'groups']

    def validate_email(self, email: str):
        view = self.context['view']
        tenant_id = view.get_tenant_id()
        if Member.objects.filter(tenant_id=tenant_id, email=email).count():
            raise ValidationError(_('This email has already been invited.'))
        return email

    def create(self, validated_data):
        request = self.context['request']
        email = validated_data['email']
        try:
            user_email = UserEmail.objects.get_by_email(email)
            validated_data['user_id'] = user_email.user_id
            if user_email.user_id == request.user.id:
                validated_data['status'] = Member.InviteStatus.ACTIVE
            else:
                validated_data['status'] = Member.InviteStatus.WAITING
        except UserEmail.DoesNotExist:
            pass
        validated_data['inviter'] = request.user
        try:
            return super().create(validated_data)
        except IntegrityError:
            raise ValidationError(
                {
                    'email': [_('This user has already been invited.')],
                }
            )


class MemberDetailSerializer(ModelSerializer):
    inviter = RelatedSerializerField(SimpleUserSerializer, required=False, read_only=True)
    role = RelatedSerializerField(RoleSerializer, required=False)
    groups = RelatedSerializerField(GroupSerializer, many=True)
    permissions = RelatedSerializerField(PermissionSerializer, many=True)

    class Meta:
        model = Member
        exclude = ['tenant', 'user']
        read_only_fields = [
            'email',
            'created_at',
        ]


class UserMembershipSerializer(ModelSerializer):
    tenant = TenantSerializer(read_only=True)
    role = RoleSerializer(read_only=True)
    groups = GroupSerializer(many=True, read_only=True)
    permissions = PermissionSerializer(many=True, read_only=True)

    class Meta:
        model = Member
        exclude = ['user', 'inviter', 'email']
        read_only_fields = [
            'inviter',
            'email',
            'created_at',
        ]
        request_include_fields = ['groups', 'role', 'permissions']

    def validate_status(self, status):
        # user agree to join the tenant
        if status != Member.InviteStatus.ACTIVE:
            raise ValidationError(_('Invalid status value.'))
        return status


class InvitationInfoSerializer(ModelSerializer):
    tenant = TenantSerializer(read_only=True)
    inviter = SimpleUserSerializer(read_only=True)

    class Meta:
        model = Member
        fields = ['tenant', 'inviter', 'email', 'status', 'created_at']



from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from .permission import Permission
from ..db import CachedManager


class RoleManager(CachedManager['Role']):
    def get_by_name(self, name: str):
        return self.get_from_cache_by_pk(name)


class Role(models.Model):
    """Pre-defined roles with permissions."""

    name = models.CharField(max_length=100, primary_key=True)
    description = models.CharField(max_length=480, blank=True)
    permissions = models.ManyToManyField(Permission, blank=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)
    objects = RoleManager()

    class Meta:
        verbose_name = _('role')
        verbose_name_plural = _('roles')
        db_table = 'saas_role'
        ordering = ['created_at']

    def __str__(self):
        return self.name



from rest_framework.mixins import ListModelMixin
from rest_framework.request import Request
from ..drf.views import Endpoint
from ..models import Role
from ..serializers.role import RoleSerializer


__all__ = ['RoleListEndpoint']


class RoleListEndpoint(ListModelMixin, Endpoint):
    serializer_class = RoleSerializer
    queryset = Role.objects.all()
    permission_classes = []
    pagination_class = None

    def get(self, request: Request, *args, **kwargs):
        """Show all supported Roles."""
        return self.list(request, *args, **kwargs)



import typing as t
from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS

if t.TYPE_CHECKING:
    from ..models import Permission, Role


def create_permissions_receiver(permissions: t.List[t.Tuple[str, str]]):
    def _create_permissions(app_config, verbosity=2, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
        if not app_config.models_module:
            return

        if app_config.label != 'saas_base':
            return

        try:
            permission_cls: t.Type['Permission'] = apps.get_model(app_config.label, 'Permission')
        except LookupError:
            return

        existed_perms = set(permission_cls.objects.values_list('name', flat=True).all())

        to_add_perms = []
        for name, description in permissions:
            if name not in existed_perms:
                to_add_perms.append(permission_cls(name=name, description=description))

        if to_add_perms:
            permission_cls.objects.using(using).bulk_create(to_add_perms, ignore_conflicts=True)
        if verbosity >= 2:
            for perm in to_add_perms:
                print(f"Adding saas_base.Permission '{perm.name}'")

    return _create_permissions


def create_roles_receiver(roles: t.List[t.Tuple[str, str, t.List[str]]]):
    def _create_roles(app_config, verbosity=2, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
        if not app_config.models_module:
            return

        if app_config.label != 'saas_base':
            return

        try:
            role_cls: t.Type['Role'] = apps.get_model(app_config.label, 'Role')
        except LookupError:
            return

        for name, description, perms in roles:
            role, created = role_cls.objects.using(using).update_or_create(
                name=name,
                defaults={'description': description},
            )
            role.permissions.set(perms)

            if created and verbosity >= 2:
                print(f"Adding saas_base.Role '{role.name}'")

    return _create_roles



from django.utils import timezone
from tests.client import FixturesTestCase
from saas_base.models import Group, Tenant


class TestGroupsAPI(FixturesTestCase):
    tenant_id = 1
    user_id = FixturesTestCase.OWNER_USER_ID

    def test_list_all_groups(self):
        self.force_login()
        resp = self.client.get('/m/groups/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]['name'], 'Admin')

    def test_create_group(self):
        self.force_login()
        data = {
            'name': 'Guest',
            'permissions': ['tenant.read'],
        }
        resp = self.client.post('/m/groups/', data=data)
        self.assertEqual(resp.status_code, 201)
        data = resp.json()
        self.assertEqual(data['name'], 'Guest')
        permission = data['permissions'][0]
        self.assertEqual(permission['name'], 'tenant.read')

    def test_expired_tenant_create_group(self):
        tenant = Tenant.objects.create(
            owner_id=self.user_id,
            name='Expired',
            slug='expired',
            expires_at=timezone.now(),
        )
        client = self.client_class()
        client.force_login(self.user)

        client.credentials(
            HTTP_X_TENANT_ID=str(tenant.id),
        )
        data = {
            'name': 'Guest',
            'permissions': ['tenant.read'],
        }
        resp = client.post('/m/groups/', data=data)
        self.assertEqual(resp.status_code, 403)
        data = resp.json()
        self.assertEqual(data['detail'], 'This tenant is expired.')

    def test_retrieve_group(self):
        self.force_login()
        group = Group.objects.filter(tenant=self.tenant_id).first()
        resp = self.client.get(f'/m/groups/{group.id}/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(len(data['permissions']), 3)

    def test_update_group_permissions(self):
        self.force_login()
        group = Group.objects.filter(tenant=self.tenant_id).first()
        data = {'permissions': ['tenant.read']}
        resp = self.client.patch(f'/m/groups/{group.id}/', data=data)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(len(data['permissions']), 1)

    def test_update_group_name(self):
        self.force_login()
        group = Group.objects.filter(tenant=self.tenant_id).first()
        data = {'name': 'Admin 2'}
        resp = self.client.patch(f'/m/groups/{group.id}/', data=data)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['name'], 'Admin 2')
        self.assertEqual(len(data['permissions']), 3)

    def test_delete_group(self):
        self.force_login()
        group = Group.objects.filter(tenant=self.tenant_id).first()
        resp = self.client.delete(f'/m/groups/{group.id}/')
        self.assertEqual(resp.status_code, 204)



from tests.client import FixturesTestCase


class TestPermissionsAPI(FixturesTestCase):
    def test_list_all_permissions(self):
        resp = self.client.get('/m/permissions/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 3)



from rest_framework import serializers
from .permission import PermissionSerializer
from ..drf.serializers import RelatedSerializerField
from ..models import Group


class GroupSerializer(serializers.ModelSerializer):
    permissions = RelatedSerializerField(PermissionSerializer, many=True)

    class Meta:
        model = Group
        fields = ['id', 'name', 'permissions', 'managed']



from rest_framework.request import Request
from rest_framework.response import Response
from ..drf.views import AuthenticatedEndpoint
from ..serializers.user import (
    UserSerializer,
    UserPasswordSerializer,
)

__all__ = [
    'UserEndpoint',
    'UserPasswordEndpoint',
]


class UserEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user', 'user:profile']
    serializer_class = UserSerializer

    def get(self, request: Request):
        """Retrieve current user information."""
        serializer: UserSerializer = self.get_serializer(request.user)
        return Response(serializer.data)

    def patch(self, request, *args, **kwargs):
        """Update current user information."""
        serializer = self.get_serializer(request.user, data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data)


class UserPasswordEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user:password']
    serializer_class = UserPasswordSerializer

    def post(self, request: Request, *args, **kwargs):
        """Update current user's password"""
        serializer: UserPasswordSerializer = self.get_serializer(request.user, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(status=204)



from __future__ import annotations

from typing import Any
from django.db import models
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.expressions import Expression
from saas_base.secure import encrypt_data, decrypt_data


class EncryptedField(models.TextField):
    description = "A field that stores bytes encrypted with JWE (X25519)"

    def get_prep_value(self, value: bytes | None) -> str | None:
        """Encrypt Python bytes into a JWE string for the database."""
        if value is None:
            return None

        if not isinstance(value, bytes):
            raise TypeError(f"EncryptedField must receive bytes, got {type(value)}")

        return encrypt_data(value)

    def get_db_prep_value(
        self,
        value: Any,
        connection: BaseDatabaseWrapper,  # noqa: ARG002
        prepared: bool = False,  # noqa: FBT001, FBT002
    ) -> Any:
        if not prepared:
            value = self.get_prep_value(value)
        return value

    def from_db_value(
        self,
        value: Any,
        expression: Expression,  # noqa: ARG002
        connection: BaseDatabaseWrapper,  # noqa: ARG002
    ) -> Any:
        return self.to_python(value)

    def to_python(self, value: str | None) -> bytes | None:
        """Decrypt a JWE string from the database back into Python bytes with Key Rotation support."""
        if value is None or not isinstance(value, str) or hasattr(self, '_already_decrypted'):
            return value
        return decrypt_data(value)

    def clean(self, value: Any, model_instance: models.Field) -> Any:
        """
        Create and assign a semaphore so that to_python method will not try
        to decrypt an already decrypted value during cleaning of a form
        """
        setattr(self, '_already_decrypted', True)
        ret = super().clean(value, model_instance)
        delattr(self, '_already_decrypted')
        return ret



import logging
from abc import ABCMeta, abstractmethod
from rest_framework.request import Request
from saas_base.drf.errors import BadRequest

logger = logging.getLogger(__name__)


class Rule(metaclass=ABCMeta):
    DEFAULT_RESPONSE_FIELD: str = 'email'
    error_message: str = 'Bad request'

    def __init__(self, **options):
        self.options = options

    def get_response_field_value(self, request: Request):
        response_field = self.options.get(
            'response_field',
            self.DEFAULT_RESPONSE_FIELD,
        )
        return request.data.get(response_field)

    @abstractmethod
    def bad_request(self, request: Request):
        pass


def check_rules(rules: list[Rule], request: Request):
    for rule in rules:
        if rule.bad_request(request):
            rule_name = rule.__class__.__name__
            logger.warning(f'Bad request: [{rule_name}].')
            raise BadRequest(detail=rule.error_message, code=rule_name)



import warnings
from saas_base.rules import Rule, BlockedEmailDomains, Turnstile, AvoidTooManyDots as TooManyDots

__all__ = [
    'Rule',
    'BlockedEmailDomains',
    'TooManyDots',
    'Turnstile',
]

warnings.warn('Use saas_base.rules instead', DeprecationWarning)
